import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import ddf.minim.analysis.*; 
import ddf.minim.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class RELEARN_microMeshMonitor extends PApplet {

// - audio


// -

String modelPath;
String audioPath;

ArrayList<HuMod> huModList;
StringList modelList;
StringList audioList;

PGraphics mesh;

boolean distort = false;

// - audio

Minim minim;  
AudioInput in;
FFT dataFFT;

float spectrumScale = 100;
int sampleRate = 1024;
int numberBands = 32;
int stepBand = numberBands/2;
int highFreqBand = numberBands-1;
int[] thresholds = new int[numberBands];
int th = stepBand;

boolean bstep = false;
boolean bcheckstep = true;
int checkNextStepMs = 330;
int justCheckedStepMs;

int totalSteps = 0;
// - 

float bassampX, bassampY, bassampZ;

boolean bbackground = true;
int backgroundCounter = 0;

public void setup(){
  
  size(800,800, P3D);
  
  huModList = new ArrayList();
  modelList = new StringList();
  audioList = new StringList();
  
  modelPath = savePath("") + "/data/models/";
  getModelList();
  println(modelList);
  initializeHuMod();

  mesh = createGraphics(width, height, P3D);
  
  audioPath = savePath("") + "/data/audio/";
  getAudioList();
  println(audioList);
  
  // - audio
  minim = new Minim(this);
  // use the getLineIn method of the Minim object to get an AudioInput
  in = minim.getLineIn();
  
  dataFFT = new FFT( in.bufferSize(), sampleRate );
  
  dataFFT.linAverages(numberBands);
  
  // thresholds init
  for(int i=0; i<numberBands; i++){
    thresholds[i] = (int)random(10, 40);
  }
  thresholds[highFreqBand] = 10;
  // - 
  
}

public void draw(){
  
  ///draw mesh////////////////////////////////////////////////////
  mesh.beginDraw();
  
	if(bbackground){
		if (distort==true){
		  mesh.background(255);
		} else {
		  mesh.background(0);
		}
	}
      mesh.pushMatrix();
    
        mesh.translate(width/2, height/2);
        
        mesh.rotateY(ROTX);
        mesh.rotateX(ROTY);
        mesh.scale(SCALE);
        
        mesh.directionalLight(255, 255, 255, 0, -1, 0);
        //mesh.ambientLight(255, 255, 255);
              
        if (distort==true){
          huModList.get(0).displayPerlinMesh();
          distort=false;
        } else {
          huModList.get(0).displayPerlinMeshBlob(bassampX, bassampY, bassampZ);
        }

    mesh.popMatrix();
  mesh.endDraw();
  
  image(mesh, 0, 0);
  ///////////////////////////////////////////////////////
  
  
  ///draw audio/////////////////////////////////////////
  rectMode(CORNERS);
  
  // - audio
  if(bcheckstep == false){
    if(millis() - justCheckedStepMs > checkNextStepMs){
      bcheckstep = true;
    }
  }
  
  // perform a forward FFT on the samples
  dataFFT.forward(in.mix);
  
  bassampZ = dataFFT.getAvg(0)*spectrumScale;
  bassampY = dataFFT.getAvg(1)*spectrumScale;
  bassampX = dataFFT.getAvg(2)*spectrumScale;
  
  // draw the full spectrum
  noFill();
  stroke(255,100);
  for(int i=0; i<dataFFT.specSize(); i++){
    line(i, height, i, height - dataFFT.getBand(i)*spectrumScale);
  }
  
  // draw the linear averages
  int w = PApplet.parseInt(width/dataFFT.avgSize());
  stroke(255);
  for(int i = 0; i < dataFFT.avgSize(); i++){
    // draw a rectangle for each average, multiply the value by spectrumScale so we can see it better
    if(stepBand == i){
      fill(255,200);
      bstep = false;
      if( (bcheckstep==true) && (dataFFT.getAvg(i)*spectrumScale > thresholds[stepBand]) ){
        bstep = true;
        distort = true;
        totalSteps++;
        justCheckedStepMs = millis();
        bcheckstep = false;
      }
    }
    else{
      fill(255,100);
    }   
    rect(i*w, height, i*w + w, height - dataFFT.getAvg(i)*spectrumScale);
    
    /*
    if(highFreqBand == i){
      if((backgroundCounter<=0) && (dataFFT.getAvg(i)*spectrumScale > thresholds[highFreqBand])){
        bbackground = false;
        backgroundCounter = 5;
      }
      if(backgroundCounter>=0){
        backgroundCounter--;
      }
      else{
        bbackground = true;
      }
    }
    */
  }
  stroke(255,0,60);
  strokeWeight(5);
  line(0,height-thresholds[stepBand], width, height-thresholds[stepBand]);
  stroke(255,100,160);
  strokeWeight(1);
  for(int i=0; i<numberBands; i++){
    line(i*w,height-thresholds[i], i*w+w, height-thresholds[i]);
  }
  
  
  float alfa = map(millis()-justCheckedStepMs, 0,checkNextStepMs, 255, 0);
  fill(255, alfa);
  noStroke();
  ellipse(10,10, 20,20);
  
  audioDrawDebugInfo(w);
  // -
    
}


public void audioDrawDebugInfo(int _w){
  fill(255,200);
  for(int i = 0; i < dataFFT.avgSize(); i++){
    text(i, 2+i*_w, height - 10);
  }
  text("thresholds[stepBand] [++|--]: " + thresholds[stepBand], 20, 20);
  text("step band [a+|s-]: " + stepBand, 20, 40);
  text("checkNextStepMs [w+|e-]: " + checkNextStepMs, 20, 60);
  text("totalSteps: " + totalSteps, 20, 80);
  text("th [v+|b-]: " + th, 20, 120);
  text("thresholds[th] [n+|m-]: " + thresholds[th], 20, 140);
  text("high freq band [d+|f-]: " + highFreqBand, 20, 160);
}


class HuMod{
  
  int index;
  String fileName;
  String[] lines;
  FloatList vList;
  
  float minZ = 1000000;
  float maxZ = -1000000;

  
  HuMod(int ind, String fp) {
    
    vList = new FloatList();
    
    index = ind;
    fileName=fp;
    getGeometry();
    sortGeometry();
    
  }
  
  
 public void displayBaseMesh(){

    mesh.fill(150);
    mesh.noStroke();
 
    mesh.beginShape(TRIANGLES);
    for (int i=0; i<vList.size(); i+=9) {
      mesh.vertex(vList.get(i+0), vList.get(i+1), vList.get(i+2)); 
      mesh.vertex(vList.get(i+3), vList.get(i+4), vList.get(i+5));      
      mesh.vertex(vList.get(i+6), vList.get(i+7), vList.get(i+8));
    }
    mesh.endShape(CLOSE);
       
  }

  
  public void displayPerlinMesh(){
   
    mesh.noFill();
    mesh.strokeWeight(1/SCALE);
    mesh.stroke(150,100);
 
    float var0, var1, var2, var3, var4, var5, var6, var7, var8;
    float amp = 0.5f;
    float freq = 0.001f;
    
    mesh.beginShape(TRIANGLES);
    for (int i=0; i<vList.size(); i+=9){
      freq = random(0.001f,0.005f);
      var0 = amp * noise(freq*millis(), vList.get(i+0));
      var1 = amp * noise(freq*millis(), vList.get(i+1));
      var2 = amp * noise(freq*millis(), vList.get(i+2));
      var3 = amp * noise(freq*millis(), vList.get(i+3));
      var4 = amp * noise(freq*millis(), vList.get(i+4));
      var5 = amp * noise(freq*millis(), vList.get(i+5));
      var6 = amp * noise(freq*millis(), vList.get(i+6));
      var7 = amp * noise(freq*millis(), vList.get(i+7));
      var8 = amp * noise(freq*millis(), vList.get(i+8));
      mesh.vertex(vList.get(i+0)+var0, vList.get(i+1)+var1, vList.get(i+2)+var2); 
      mesh.vertex(vList.get(i+3)+var3, vList.get(i+4)+var4, vList.get(i+5)+var5);      
      mesh.vertex(vList.get(i+6)+var6, vList.get(i+7)+var7, vList.get(i+8)+var8);
    }
    mesh.endShape(CLOSE);
    
    //println("mesh");
    
  }
  
  public void displayPerlinMeshBlob(float _bassampX, float _bassampY, float _bassampZ){
   
    mesh.fill(200);
    //mesh.noStroke();
    //mesh.noFill();
    mesh.strokeWeight(1/SCALE);
    mesh.stroke(150,50);
 
    float var0, var1, var2, var3, var4, var5, var6, var7, var8;
    float ampX = map(_bassampX, 0, 200, 0.1f, 0.5f);
    float ampY = map(_bassampY, 0, 200, 0.1f, 0.5f);
    float ampZ = map(_bassampZ, 0, 200, 0.1f, 0.5f);
    float freq = 0.001f;
    
    mesh.beginShape(TRIANGLES);
    for (int i=0; i<vList.size(); i+=9){
      //freq = random(0.001,0.005);
      var0 = ampX * noise(freq*millis(), vList.get(i+0));
      var1 = ampY * noise(freq*millis(), vList.get(i+1));
      var2 = ampZ * noise(freq*millis(), vList.get(i+2));
      var3 = ampX * noise(freq*millis(), vList.get(i+3));
      var4 = ampY * noise(freq*millis(), vList.get(i+4));
      var5 = ampZ * noise(freq*millis(), vList.get(i+5));
      var6 = ampX * noise(freq*millis(), vList.get(i+6));
      var7 = ampY * noise(freq*millis(), vList.get(i+7));
      var8 = ampZ * noise(freq*millis(), vList.get(i+8));
      mesh.vertex(vList.get(i+0)+var0, vList.get(i+1)+var1, vList.get(i+2)+var2); 
      mesh.vertex(vList.get(i+3)+var3, vList.get(i+4)+var4, vList.get(i+5)+var5);      
      mesh.vertex(vList.get(i+6)+var6, vList.get(i+7)+var7, vList.get(i+8)+var8);
    }
    mesh.endShape(CLOSE);
        
  }
  
  public void sortGeometry(){
    
    int lineCount=0;
    String tempVert0[];
    String tempVert1[];
    String tempVert2[];
    
    while (lineCount<lines.length-3){
      
      lines[lineCount] = trim(lines[lineCount]);
      tempVert0 = split(lines[lineCount], " ");  
      
      if (tempVert0[0].equals("vertex")) {      
        
        tempVert1 = split(lines[lineCount+1], " ");
        tempVert2 = split(lines[lineCount+2], " ");  
           
        vList.append(PApplet.parseFloat(tempVert0[1]));
        vList.append(PApplet.parseFloat(tempVert0[2]));
        vList.append(PApplet.parseFloat(tempVert0[3]));
        vList.append(PApplet.parseFloat(tempVert1[1]));
        vList.append(PApplet.parseFloat(tempVert1[2]));
        vList.append(PApplet.parseFloat(tempVert1[3]));
        vList.append(PApplet.parseFloat(tempVert2[1]));
        vList.append(PApplet.parseFloat(tempVert2[2]));
        vList.append(PApplet.parseFloat(tempVert2[3]));
        
        lineCount+=3;
        
      } else {
        lineCount++;
      }
    }
    
   ///get min/ max z value
   for (int i = 1; i<vList.size(); i+=3){
     if (vList.get(i)<minZ){ minZ = vList.get(i);}
     if (vList.get(i)>maxZ){ maxZ = vList.get(i);}
   }
   
   //println(minZ + " - " + maxZ);
   
   ///move model in z to midpoint
   for (int i = 1; i<vList.size(); i+=3){
     //println(vList.get(i));
     float tPt =  vList.get(i);
     vList.set(i, tPt - ((maxZ - minZ)/2));
   }
  
  }

  public void getGeometry(){
    lines = loadStrings(modelPath + fileName);   
  }
  
}

public void initializeHuMod(){
  
  for (int i=0; i<modelList.size(); i++){
    huModList.add(new HuMod(i, modelList.get(i)));
  }

}
public void getModelList(){
    
  File folder = new File(modelPath);    
  File[] listOfFiles = folder.listFiles();

  for (File file : listOfFiles) {
    if (file.isFile()) {
      modelList.append(file.getName());
    }
  }
  
  modelList.sort();
}


public void getAudioList(){
    
  File folder = new File(audioPath);    
  File[] listOfFiles = folder.listFiles();

  for (File file : listOfFiles) {
    if (file.isFile()) {
      audioList.append(file.getName());
    }
  }
  
  audioList.sort();
  
}



float ROTX = PI;                        
float ROTY = PI;
float SCALE = 80;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////control the rotation and scale display of the model using the mouse
public void mouseDragged() {
   
   if (mouseButton == LEFT) {
     ROTX -= (mouseX - pmouseX) * 0.01f;
     ROTY -= (mouseY - pmouseY) * 0.01f;
   }
   if (mouseButton == RIGHT) {
     if (mouseY>pmouseY){
        SCALE = SCALE+(mouseY - pmouseY);
     } else {
        SCALE = SCALE+(mouseY - pmouseY);
     }
     SCALE = abs(SCALE);
   }
   redraw();
}

public void keyReleased(){

  
}

//////////////////audio controls

public void keyPressed(){
  if(key == '+'){
    thresholds[stepBand]++;
  }
  else if(key == '-'){
    thresholds[stepBand]--;
  }
  if(key == 'v'){
    th++;
    th = min(th, numberBands-1);
  }
  else if(key == 'b'){
    th--;
    th = max(th,0);
  }
  if(key == 'n'){
    thresholds[th]++;
  }
  else if(key == 'm'){
    thresholds[th]--;
  }
  else if(key == 'a'){
    stepBand++;
    stepBand = min(stepBand, numberBands);
  }
  else if(key == 's'){
    stepBand--;
    stepBand = max(stepBand,0);
  }
  else if(key == 'w'){
    checkNextStepMs += 10;
  }
  else if(key == 'e'){
    checkNextStepMs -= 10;
  }
  else if(key == 'x'){
    bbackground = !bbackground;
  }
  else if(key == 'd'){
    highFreqBand++;
    highFreqBand = min(highFreqBand, numberBands);
  }
  else if(key == 'f'){
    highFreqBand--;
    highFreqBand = max(highFreqBand,0);
  }
}







  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "RELEARN_microMeshMonitor" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
